import os.path
import stat
import collections

import numpy as np
import h5py

import util.io.fs
import util.io.universal
import util.logging
logger = util.logging.logger


## Options File

class OptionsFile():

    def __init__(self, file, mode='a', replace_environment_vars_at_set=False, replace_environment_vars_at_get=False):
        ## prepare file name
        if os.path.isdir(file):
            file = os.path.join(file, 'options.hdf5')
        else:
            (root, ext) = os.path.splitext(file)
            if ext == '':
                file += '.hdf5'
        ## open
        self.open(file, mode)
        ## save replace variable
        self.replace_environment_vars_at_set = replace_environment_vars_at_set
        self.replace_environment_vars_at_get = replace_environment_vars_at_get


    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


    def __setitem__(self, key, value):
        ## check value
        if value is None:
            raise ValueError('Value None is not allowed (for key {}).'.format(key))
        
        ## replace env
        if self.replace_environment_vars_at_set:
            value = self._replace_environment_vars(value)

        ## check if writable
        if not self.is_writable():
            raise OSError('Option file {} is not writable.'.format(self.filename))
        
        ## insert supported types
        try:
            f = self.__hdf5_file
            ## set if key exists
            try:
                f[key][()] = value
            ## set if key not exists
            except KeyError:
                f[key] = value
        
        ## try to insert unsupported types
        except TypeError as e:
            successfully_inserted = False
            
            ## if dict, insert each item since dict is not supported in HDF5
            if isinstance(value, dict):
                for (key_i, value_i) in value.items():
                    self['{}/{}'.format(key, key_i)] = value_i
                successfully_inserted = True
            
            ## if iterable of unicode strings
            if not successfully_inserted:
                value_array = np.asanyarray(value)
                if value_array.dtype.kind == 'U':                
                    h5_dtype = h5py.special_dtype(vlen=str)
                    dataset = f.create_dataset(key, value_array.shape, dtype=h5_dtype)
                    dataset[:] = value_array
                    successfully_inserted = True
                    
            ## if tuple or list with different types, insert each item since generic object arrays are not supported in HDF5
            if not successfully_inserted and (isinstance(value, tuple) or isinstance(value, list)):
                for i in range(len(value)):
                    self['{}/{}'.format(key, i)] = value[i]
                successfully_inserted = True
            
            ## if not insertable, raise error
            if not successfully_inserted:
                raise TypeError('The value {} for key {} could not be stored, because this type of value is not supported in hdf5.'.format(value, key)) from e
    

    def __getitem__(self, name):
        ## get value
        try:
            item = self.__hdf5_file[name]
        except KeyError as e:
            raise KeyError('The key {} is not in the option file {}.'.format(name, self.filename)) from e
        value = item.value
        ## replace env
        if self.replace_environment_vars_at_get:
            value = self._replace_environment_vars(value)
        ## return
        return value


    def __delitem__(self, name):
        try:
            del self.__hdf5_file[name]
        except KeyError as e:
            raise KeyError('The key {} is not in the option file {}.'.format(name, self.filename)) from e
        #TODO delete empty groups
    

    @property
    def __hdf5_file(self):
        hdf5_file_object = self.__hdf5_file_object
        if hdf5_file_object is not None:
            return hdf5_file_object
        else:
            raise ValueError('File is closed.')


    def __str__(self):
        return 'Option file {}'.format(self.filename)

    @property
    def filename(self):
        hdf5_file = self.__hdf5_file
        return hdf5_file.filename


    def open(self, file, mode='a'):
        self.close()

        logger.debug('Opening option file {} with mode {}.'.format(file, mode))

        try:
            f = h5py.File(file, mode=mode)
        except OSError as e:
            if mode != 'r':
                logger.debug('File {} could not been open. Trying read_only mode.'.format(file))
                self.open(file, mode='r')
            else:
                raise OSError('Option file {} could not been open. Error: {}. Error code: {}.'.format(file, e.strerror, e.errno)) from e
        else:
            logger.debug('File {} opened.'.format(file))
            self.__hdf5_file_object = f


    def close(self):
        try:
            file = self.__hdf5_file_object
        except AttributeError:
            file = None

        if file is not None:
            logger.debug('Closing option file {}.'.format(file.filename))
            file.close()
            self.__hdf5_file_object = None


    @staticmethod
    def _replace_environment_vars(value):
        def is_str(value):
            return isinstance(value, str) or isinstance(value, bytes)
        
        def expand(value):
            if is_str(value):
                value = os.path.expanduser(value)
                value = os.path.expandvars(value)
            return value

        if is_str(value):
            value = expand(value)
        elif isinstance(value, collections.Iterable) and any(map(is_str, value)):
            value = tuple(map(expand, value))
        return value


    ## permissions

    def is_writable(self):
        f = self.__hdf5_file
        return f.mode == 'r+'

    def is_read_only(self):
        f = self.__hdf5_file
        return f.mode == 'r'

    def make_writable(self):
        if not self.is_writable():
            logger.debug('Opening {} writable.'.format(self.filename))
            file = self.filename
            self.close()
            util.io.fs.make_writable(file)
            self.open(file)
        else:
            logger.debug('File {} is writable.'.format(self.filename))
        assert self.is_writable

    def make_read_only(self):
        if not self.is_read_only():
            logger.debug('Opening {} read_only.'.format(self.filename))
            file = self.filename
            self.close()
            util.io.fs.make_read_only(file)
            self.open(file, 'r')
        else:
            logger.debug('File {} is read_only.'.format(self.filename))
        assert self.is_read_only


    ## print

    def print_all_options(self):
        f = self.__hdf5_file

        def print_option(name, object):
            ## check if dataset
            try:
                value = object.value
            except AttributeError:
                value = None

            ## check type
            if value is not None:
                print('{}: {}'.format(name, value))

        f.visititems(print_option)


    ## replace str

    def get_all_str_options(self):
        string_object_list = []

        def append_if_string_option(name, object):
            ## check if dataset
            try:
                value = object.value
            except AttributeError:
                pass
            ## check type
            else:
                if isinstance(value, str) or (isinstance(value, np.ndarray) and any(tuple(map(lambda v: isinstance(v[np.newaxis][0], str), np.nditer(value, flags=['refs_ok']))))):
                    string_object_list.append(name)

        f = self.__hdf5_file
        f.visititems(append_if_string_option)

        return string_object_list
    

    def replace_all_str_options(self, old_str, new_str):
        def replace_string_in_option(option, object):
            ## check if dataset
            try:
                value = object.value
            except AttributeError:
                pass
            ## check type
            else:
                ## replace in (scalar) string
                if isinstance(value, str):
                    new_value = value.replace(old_str, new_str)
                ## replace in string array
                elif isinstance(value, np.ndarray) and all(map(lambda v: isinstance(v[np.newaxis][0], str), np.nditer(value, flags=['refs_ok']))):
                    new_value = value.copy()
                    for v in np.nditer(new_value, flags=['refs_ok'], op_flags=['readwrite']):
                        v[...] = v[np.newaxis][0].replace(old_str, new_str)
                    new_value = new_value.astype('U')
                ## otherwise skip
                else:
                    new_value = None
                ## set replaced option value
                if new_value is not None and np.any(value != new_value):
                    self[option] = new_value
                    logger.info('Option {} updated from {} to {}.'.format(option, value, new_value))
                else:
                    logger.debug('Option {} with value {} does not have to be updated.'.format(option, value))

        f = self.__hdf5_file
        f.visititems(replace_string_in_option)




## Options Dict

class OptionError(KeyError):
    
    def __init__(self, option, option_object, message):
        self.option = option
        self.option_object = option_object
        message = '{}: {}'.format(type(option_object).__name__, message)
        super().__init__(message)


class UnknownOptionError(OptionError):
    
    def __init__(self, option, option_object):
        message = 'Option {} is unknown.'.format(option)
        super().__init__(option, option_object, message)


class NoneSetOptionError(OptionError):
    
    def __init__(self, option, option_object):
        message = 'Option {} is not set.'.format(option)
        super().__init__(option, option_object, message)


class ImmutableOptionError(OptionError):
    
    def __init__(self, option, option_object):
        message = 'Option {} can not be changed.'.format(option)
        super().__init__(option, option_object, message)


class IncalculableOptionError(OptionError):
    
    def __init__(self, option, option_object):
        message = 'Option {} is incalculable.'.format(option)
        super().__init__(option, option_object, message)


class UnknownListenerError(KeyError):
    
    def __init__(self, listener):
        message = 'Listener {} is unknown.'.format(listener)
        super().__init__(message)



class OptionsBase():
    
    NON_OPTION_NAMES = ('_options', '_option_names')
    
    def __init__(self, options=None, default_options=None, option_names=None):
        logger.debug('Initiating {} with options {}, default_options {} and option_names {}.'.format(type(self).__name__, options, default_options, option_names))
        self._option_names = option_names
        self._options = {}

        ## set passed and default options
        if default_options is not None:
            for option in default_options:
                if options is None or option not in options:
                    self[option] = default_options[option]

        ## set options
        if options is not None:
            for option in options:
                self[option] = options[option]
    
    
    ## dict methods
    
    def __getattr__(self, option):
        if self._has_option(option):
            return self._get_option(option)
        else:
            return super().__getattribute__(option)    
    
    
    def __setattr__(self, option, value):
        if self._has_option(option):
            self._set_option(option, value)
        else:
            object.__setattr__(self, option, value)

    
    def __delattr__(self, option):
        if self._has_option(option):
            self._del_option(option)
        else:
            object.__delattr__(self, option)

    
    def __getitem__(self, option):        
        return eval('self.'+option)

    
    def __setitem__(self, option, value):
        self.__setattr__(option, value)
    
    
    def __delitem__(self, option):
        self.__delattr__(option)
    

    def __len__(self):
        return len(self._options)    
    
    
    def __contains__(self, option):
        return self._has_value(option)
    
    
    def __iter__(self):
        return self._options.keys()
    
    
    def __str__(self):
        return str(self._options)
    
    
    def __repr__(self):
        return '{module_name}.{class_name}({options!r})'.format(module_name=self.__module__, class_name=self.__class__.__name__,  options=self._options)
    
    
    ## get and set options method
    
    def _get_option(self, option):
        logger.debug('Getting option {}.'.format(option))
        try:
            return self._options[option]
        except KeyError:
            logger.debug('Option {} is not set.'.format(option))
            raise NoneSetOptionError(option, self)
    
    
    def _set_option(self, option, value):
        logger.debug('Setting option {}.'.format(option))
        self._options[option] = value
    
    
    def _del_option(self, option, not_exist_okay=False):
        logger.debug('Deleting option {} with not_exist_okay {}.'.format(option, not_exist_okay))
        try:
            del self._options[option]
        except KeyError:
            logger.debug('Option {} is not set.'.format(option))
            if not not_exist_okay:
                raise NoneSetOptionError(option, self)
    
    
    def _clear_all_options(self):
        for option in tuple(self._options.keys()):
            self._del_option(option)
    
    
    def _has_option(self, option):
        return (not option in type(self).NON_OPTION_NAMES) and (self._option_names is None or option in self._option_names)
        
    
    def _has_value(self, option):
        return option in self._options



class OptionsWithValueCheck(OptionsBase):
    
    VALUE_CHECK_METHOD_NAME = '{}_check'
    
    
    def _set_option(self, option, value):
        value_check_method_name = self.VALUE_CHECK_METHOD_NAME.format(option)

        try:
            value_check_method = getattr(self, value_check_method_name)
        except AttributeError:
            pass
        else:
            checked_value = value_check_method(value)
            if checked_value is not None:
                value = checked_value
        
        super(). _set_option(option, value)
    


class OptionsWithListeners(OptionsBase):
    
    NON_OPTION_NAMES = OptionsBase.NON_OPTION_NAMES + ('_listeners',)

    def __init__(self, **kargs):
        self._listeners = {}
        super().__init__(**kargs)
    
    
    ## get and set options method
    
    def _set_option(self, option, new_value):
        
        ## check old value
        if self._has_value(option):
            old_value = self._get_option(option)
            must_set = np.any(new_value != old_value)
        else:
            must_set = True
            
        ## set new option value and call listener
        if must_set:
            super()._set_option(option, new_value)
            self._call_listeners(option)
    
    
    def _del_option(self, option, not_exist_okay=False):
        super()._del_option(option, not_exist_okay=not_exist_okay)
        self._call_listeners(option)
    
    
    ## listener methods
    
    def add_listener(self, option, listener):
        ## check input
        if not self._has_option(option):
            raise UnknownOptionError(option, self)
        if not callable(listener):
            raise ValueError('The listener must be callable, but it is {}.'.format(listener))
        
        ## add listener
        try:
            listener_list = self._listeners[option]
        except KeyError:
            self._listeners[option] = [listener]
        else:
            listener_list.append(listener)
    
    
    def remove_listener(self, option, listener):
        if not self._has_option(option):
            raise UnknownOptionError(option, self)
        if not callable(listener):
            raise ValueError('The listener must be callable, but it is {}.'.format(listener))
            
        try:
            listener_list = self._listeners[option]
        except KeyError:
            raise UnknownOptionError(listener, self)
        else:
            listener_list.remove(listener)
    
    
    def get_listeners(self, option):
        try:
            return self._listeners[option]
        except KeyError:
            return []
    
    
    def _call_listeners(self, option):
        try:
            listener_list = self._listeners[option]
        except KeyError:
            pass
        else:
            if self._has_value(option):
                value =  self._get_option(option)
            else:
                value = None
            
            for listener in listener_list:
                listener(option, value)



class OptionsWithListenersAndCopy(OptionsWithListeners):
    
    def add_copy_listener(self, independent_option, dependent_option, dependent_option_object=None, copy_now=True):
        if dependent_option_object is None:
            dependent_option_object = self
        
        def listener(independent_option, new_value):
            dependent_option_object[dependent_option] = new_value
        
        if copy_now:
            try:
                value = self[independent_option]
            except NoneSetOptionError:
                pass
            else:
                listener(independent_option, value)
        
        self.add_listener(independent_option, listener)



class OptionsWithListenersAndDependencies(OptionsWithListeners):
    
    DEPENDENT_VALUE_CHANGED_METHOD_NAME = '{}_depending_value_changed'

    def __init__(self, dependencies=None, **kargs):
        super().__init__(**kargs)
        
        if dependencies is not None:
            for dependent_option, independent_option in dependencies.items():
                self.add_dependency(independent_option, dependent_option)
    
    
    def add_dependency(self, independent_option, dependent_option, dependent_option_object=None):
        if dependent_option_object is None:
            dependent_option_object = self

        depending_value_changed_method_name = self.DEPENDENT_VALUE_CHANGED_METHOD_NAME.format(dependent_option)
        
        def listener(independent_option, new_value):
            try:
                depending_value_changed_method = getattr(dependent_option_object, depending_value_changed_method_name)
            except AttributeError:
                remove_value = True
            else:
                remove_value = depending_value_changed_method(independent_option, new_value)
            
            if remove_value:
                dependent_option_object._del_option(dependent_option, not_exist_okay=True)
        
        self.add_listener(independent_option, listener)
    


class OptionsWithCache(OptionsBase):
    
    CALCULATION_METHOD_NAME= '{}_calculated'
    
    def _get_option(self, option):
        try:
            return super()._get_option(option)
        except NoneSetOptionError as e:
            try:
                return self._recalculated_option_value(option)
            except IncalculableOptionError:
                raise e
        
    
    def _recalculated_option_value(self, option):
        logger.debug('Recalculating option {}.'.format(option))
        calculation_method_name = self.CALCULATION_METHOD_NAME.format(option)
        try:
            value = getattr(self, calculation_method_name)
        except (AttributeError, NoneSetOptionError):
            raise IncalculableOptionError(option, self)
        else:
            self._set_option(option, value)
            return value



class Options(OptionsWithCache, OptionsWithValueCheck, OptionsWithListenersAndCopy, OptionsWithListenersAndDependencies):
    pass
    
    

def as_options(option_object, option_class=Options):
    if not isinstance(option_object, option_class):
        option_object = option_class(option_object)
    return option_object




import abc

import numpy as np

import util.logging
logger = util.logging.logger



class Database:
    
    def __init__(self, value_reliable_decimal_places=18, tolerance_options=None):
        
        ## set value format
        value_reliable_decimal_places = int(value_reliable_decimal_places)
        assert value_reliable_decimal_places >= 0
        self.value_fmt = '%.{}f'.format(value_reliable_decimal_places)

        ## set tolerance options
        if tolerance_options is None:
            tolerance_options = {}
        self._tolerance_options = {}
        
        ## set relative option
        try:
            relative = tolerance_options['relative']
        except KeyError:
            relative = None
        
        if relative is None:
            relative = np.array([0])
        else:
            relative = np.asanyarray(relative).reshape(-1)
            if np.any(relative < 0):
                raise ValueError('The relative tolerance {} has to be positive.'.format(relative))
        
        self._tolerance_options['relative'] = relative

        ## min absolute tolerance
        min_absolute_tolerance = 10**(-value_reliable_decimal_places)

        ## set absolute option
        try:
            absolute = tolerance_options['absolute']
        except KeyError:
            absolute = None
        
        if absolute is None:
            absolute = np.asarray([min_absolute_tolerance])
        else:
            absolute = np.asanyarray(absolute).reshape(-1)
            if np.any(absolute < 0):
                raise ValueError('The absolute tolerance {} has to be positive.'.format(absolute))
            elif np.any(absolute < min_absolute_tolerance):
                logger.warn('The absolute tolerance {} is not support. Using smallest supported absolute tolerance {}.'.format(absolute, min_absolute_tolerance))
                absolute = np.asanyarray(absolute, dtype=np.float64)
                absolute[absolute < min_absolute_tolerance] = min_absolute_tolerance
        
        self._tolerance_options['absolute'] = absolute
        
        ## check both options
        if not (len(self._tolerance_options['absolute']) == 1 or len(self._tolerance_options['relative']) == 1 or len(self._tolerance_options['relative']) == len(self._tolerance_options['absolute'])):
            raise ValueError('The relative and absolute tolerances habe to be scalaras or arrays of equal length, but the relative tolerance is {} and the absolute is {}.'.format(self._tolerance_options['relative'], self._tolerance_options['absolute']))
        
        logger.debug('Index database initiated with {} value format and tolerance options {}.'.format(self.value_fmt, self._tolerance_options))
        
    
    ## tolerances
    
    @property
    def relative_tolerance(self):
        return self._tolerance_options['relative']  
    
    @property
    def absolute_tolerance(self):
        return self._tolerance_options['absolute']
    

    ## value comparison
    
    def value_difference(self, v1, v2):
        ## check input
        if len(v1) != len(v2):
            raise ValueError('Both values must have equal lengths, but length of {} is {} and length of {} is {}.'.format(v1, len(v1), v2, len(v2)))
        if not len(self.relative_tolerance) in (1, len(v1)):
            raise ValueError('The relative tolerances must be a scalar or of equal length as the values, but the relative tolerance is {} with length {} and the values have length {}.'.format(self.relative_tolerance, len(self.relative_tolerance), len(v1)))
        if not len(self.relative_tolerance) in (1, len(v1)):
            raise ValueError('The absolute tolerances must be a scalar or of equal length as the values, but the absolute tolerance is {} with length {} and the values have length {}.'.format(self.absolute_tolerance, len(self.absolute_tolerance), len(v1)))
            
        
        ## calculate value weights
        relative_weights = np.minimum(np.abs(v1), np.abs(v2))

        assert len(self.relative_tolerance) in (1, len(v1))
        assert len(self.absolute_tolerance) in (1, len(v1))
        total_weights = np.maximum(relative_weights * self.relative_tolerance, self.absolute_tolerance)
        assert np.all(total_weights > 0)
        
        ## calculate max difference
        value_differences = np.abs(v1 - v2) / total_weights
        value_difference = value_differences.max()

        return value_difference

    
    def are_values_equal(self, v1, v2):
        return self.value_difference(v1, v2) <= 1
    
    
    ## access

    @abc.abstractmethod
    def get_value(self, index):
        raise NotImplementedError()

    def has_value(self, index):
        try:
            value = self.get_value(index)
        except DatabaseIndexError:
            has_value = False
        else:
            has_value = True

        logger.debug('{}: Has value at index {}: {}.'.format(self, index, has_value))
        return has_value

    @abc.abstractmethod
    def set_value(self, index, value, overwrite=True):
        raise NotImplementedError()

    def add_value(self, value):
        logger.debug('{}: Adding value {}'.format(self, value))
        
        ## get used indices
        used_indices = self.used_indices()
        
        ## create value
        index = 0
        created = False
        while not created:
            if not index in used_indices:
                try:
                    self.set_value(index, value, overwrite=False)
                except DatabaseOverwriteError:
                    index += 1
                else:
                    created = True
            else:
                index += 1
        
        ## return index
        logger.debug('{}: Value {} added with index {}.'.format(self, value, index))
        return index


    @abc.abstractmethod
    def used_indices(self):
        raise NotImplementedError()

    def number_of_used_indices(self):
        return len(self.used_indices())  

    @abc.abstractmethod
    def remove_index(self, index):
        raise NotImplementedError()


    def closest_indices(self, value):
        logger.debug('{}: Calculating closest indices for value {}.'.format(self, value))
        value = np.asanyarray(value)
        
        ## get all used indices
        used_indices = self.used_indices()
        used_indices = np.asarray(used_indices)
        
        ## init value differences
        n = len(used_indices)
        value_differences = np.empty(n)

        ## calculate value differences
        for i in range(n):
            current_index = used_indices[i]
            try:
                current_value = self.get_value(current_index)
            except DatabaseIndexError as e:
                logger.warnig('{}: Could not read the value file for index {}: {}'.format(self, current_index, e.with_traceback(None)))
                value_differences[i] = float('inf')
            else:
                value_differences[i] = self.value_difference(value, current_value)
        
        ## return sorted indices
        sort = np.argsort(value_differences)
        return used_indices[sort]


    def closest_index(self, value):
        logger.debug('{}: Searching for index of value as close as possible to {}.'.format(self, value))
        
        ## get closest indices
        closest_indices = self.closest_indices(value)
        
        ## return
        if len(closest_indices) > 0:
            logger.debug('{}: Closest index is {}.'.format(self, closest_indices[0]))
            return closest_indices[0]
        else:
            logger.debug('{}: No closest index found.'.format(self))
            return None


    def index(self, value):
        ## search for directories with matching parameters
        logger.debug('{}: Searching for index of value {}.'.format(self, value))

        closest_index = self.closest_index(value)
        if closest_index is not None and self.are_values_equal(value, self.get_value(closest_index)):
            logger.debug('{}: Index for value {} is {}.'.format(self, value, closest_index))
            return closest_index
        else:
            logger.debug('{}: No index found for value {}.'.format(self, value))
            return None


    def get_or_add_index(self, value, add=True):
        index = self.index(value)
        if index is None and add:
            index = self.add_value(value)
        return index



class DatabaseIndexError(IndexError):
    
    def __init__(self, index):
        message = 'Database has no value at index {}.'.format(index)
        super().__init__(message)


class DatabaseOverwriteError(Exception):    

    def __init__(self, index):
        message = 'Database has value at index {}. Overwrite is not permitted.'.format(index)
        super().__init__(message)
    
        


import warnings

import numpy as np
import scipy.sparse

import util.math.matrix
import util.math.sparse.check

import util.logging
logger = util.logging.logger



def forward_substitution(L, b, dtype=np.float64):
    logger.debug('Starting forward substitution for system of shape {}'.format(b.shape))

    ## check input
    L = util.math.sparse.check.sorted_squared_csr(L)
    if b.ndim not in [1, 2]:
        raise ValueError('b must have 1 or 2 dims but its shape is {}.'.format(b.shape))
    if L.shape[1] != b.shape[0]:
        raise ValueError('The size of the second dim of L must be equal to the size of the first dim of b but the shape of L is {} and the shape of b is {}.'.format(L.shape, b.shape))
    if dtype is None:
        dtype = np.find_common_type([L.dtype, b.dtype, np.float32], [])

    ## init
    b = np.asarray_chkfinite(b)
    x = np.zeros(b.shape, dtype=dtype)
    column_start = L.indptr[0]

    ## fill x (forward)
    for i in range(len(b)):
        column_stop = L.indptr[i+1]

        ## check regularity and triangularity
        if column_stop <= column_start:
            raise util.math.matrix.SingularMatrixError(L, 'The {}th row is zero!'.format(i))
        if L.indices[column_stop-1] > i:
            raise util.math.matrix.NoLeftTriangularMatrixError(L, 'The entry at ({},{}) is not zero!'.format(i, L.indices[column_stop-1]))
        if L.indices[column_stop-1] < i:
            raise util.math.matrix.SingularMatrixError(L, 'The {}th diagonal entry of the tridiagonal matrix is zero!'.format(i))

        ## compute value
        column_indices = L.indices[column_start:column_stop-1]    # skip diagonal entry
        assert np.all(column_indices[:-1] < column_indices[1:])
        data = L.data[column_start:column_stop-1]

        for j, Lij in zip(column_indices, data):
            x[i] -= Lij * x[j]
            assert j < i
        x[i] += b[i]
        x[i] /= L.data[column_stop-1]       # divide by ith diagonal entry

        ## next row
        column_start = column_stop

    ## return
    return x




def backward_substitution(U, b, dtype=np.float64):
    logger.debug('Starting backward substitution for system of shape {}'.format(b.shape))

    ## check input
    U = util.math.sparse.check.sorted_squared_csr(U)
    if b.ndim not in [1, 2]:
        raise ValueError('b must have 1 or 2 dims but its shape is {}.'.format(b.shape))
    if U.shape[1] != b.shape[0]:
        raise ValueError('The size of the second dim of U must be equal to the size of the first dim of b but the shape of U is {} and the shape of b is {}.'.format(U.shape, b.shape))
    if dtype is None:
        dtype = np.find_common_type([U.dtype, b.dtype, np.float32], [])

    ## init
    b = np.asarray_chkfinite(b)
    x = np.zeros(b.shape)
    column_stop = U.indptr[len(b)]

    ## fill x (backward)
    for i in range(len(b)-1, -1, -1):
        column_start = U.indptr[i]

        ## check regularity and triangularity
        if column_stop <= column_start:
            raise util.math.matrix.SingularMatrixError(U, 'The {}th row is zero!'.format(i))
        if U.indices[column_start] < i:
            raise util.math.matrix.NoRightTriangularMatrixError(U, 'The entry at ({},{}) is not zero!'.format(i, U.indices[column_start]))
        if U.indices[column_start] > i:
            raise util.math.matrix.SingularMatrixError(U, 'The {}th diagonal entry of the tridiagonal matrix is zero!'.format(i))

        ## compute value
        column_indices = U.indices[column_start+1:column_stop]    # skip diagonal entry
        assert np.all(column_indices[:-1] < column_indices[1:])
        data = U.data[column_start+1:column_stop]

        for j, Uij in zip(column_indices, data):
            x[i] -= Uij * x[j]
            assert j > i
        x[i] += b[i]
        x[i] /= U.data[column_start]       # divide by ith diagonal entry

        ## next row
        column_stop = column_start

    ## return
    return x




def LU(L, U, b, P=None, dtype=np.float64):
    logger.debug('Solving system of dim {} with LU factors'.format(len(b)))

    if P is not None:
        util.math.sparse.check.permutation_matrix(P)
        b = P * b

    x = forward_substitution(L, b, dtype=dtype)
    x = backward_substitution(U, x, dtype=dtype)

    if P is not None:
        x = P.transpose() * x

    return x



def cholesky(L, b, P=None, dtype=np.float64):
    '''
    P A P' = L L'
    '''

    logger.debug('Solving system of dim {} with cholesky factors'.format(len(b)))

    ## convert L and U to csr format
    is_csr = scipy.sparse.isspmatrix_csr(L)
    is_csc = scipy.sparse.isspmatrix_csc(L)

    if not is_csr and not is_csc:
        warnings.warn('cholesky requires L be in CSR or CSC matrix format. Converting matrix.', scipy.sparse.SparseEfficiencyWarning)

    if is_csc:
        U = L.transpose()
    if not is_csr:
        L = L.tocsr()
    if not is_csc:
        U = L.transpose().tocsr()

    assert scipy.sparse.isspmatrix_csr(L)
    assert scipy.sparse.isspmatrix_csr(U)

    ## compute
    return LU(L, U, b, P=P, dtype=dtype)




import numpy as np

import util.logging
logger = util.logging.logger


def max_dtype(a, b):
    logger.debug('Calculating max dtype of {} and {}.'.format(a,b))

    if isinstance(a, np.floating):
        if isinstance(b, np.integer):
            return a
        if isinstance(b, np.floating):
            if np.ffinfo(a).resolution > np.ffinfo(b).resolution:
                return a
            else:
                return b


    if isinstance(a, np.integer):
        if isinstance(b, np.integer):
            if np.iinfo(a).max > np.iinfo(b).max:
                return a
            else:
                return b
        if isinstance(b, np.floating):
            return b

    raise ValueError('Dtype {} and {} are not comparable.'.format(a,b))



def min_int_dtype(*values, unsigned=False):    
    ## get max value
    if len(values) > 0:
        max_value = max(values)
    else:
        max_value = 0

    ## get dtype
    if unsigned:
        int_dtypes = (np.uint8, np.uint16, np.uint32, np.uint64)
    else:
        int_dtypes = (np.int8, np.int16, np.int32, np.int64)
    i = 0
    while np.iinfo(int_dtypes[i]).max < max_value:
        i = i+1
    dtype = int_dtypes[i]
    
    ## return 
    logger.debug('Minimal int dtype (unsigned={}) for values {} is {}.'.format(unsigned, values, dtype))
    return dtype




import argparse

import util.math.sparse.decompose.with_cholmod
import util.io.object

import util.logging

if __name__ == "__main__":
    ## configure arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('load_file', type=argparse.FileType('r'))
    parser.add_argument('save_file', type=argparse.FileType('w'),)
    parser.add_argument('-d', '--debug', action='store_true', help='Print debug infos.')
    args = parser.parse_args()

    with util.logging.Logger(disp_stdout=args.debug):
        A = util.io.object.load(args.load_file.name)
        B = util.math.sparse.decompose.with_cholmod.approximate_positive_definite(A, min_abs_value=0.01)
        util.io.object.save(args.save_file.name, B)